/**
  *  \file client/dialogs/cargohist.cc
  *  \brief Cargo History Dialog
  *
  *  This is a straight-forward port of histscan.pas::ShowLastKnownCargo.
  *  Instead of a custom data structure with a custom display routine,
  *  it generates its report directly into a RichDocument.
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "client/dialogs/cargohist.h"
#include "ui/richdocument.h"
#include "ui/uiglobal.h"
#include "util/nls.h"
#include "game/cargo.h"
#include "util/misc.h"
#include "game/ship.h"
#include "ui/layout.h"
#include "ui/window.h"
#include "gfx/palette.h"
#include "ui/msgbox.h"
#include "ui/widgets/button.h"
#include "ui/widgets/keytranslator.h"
#include "game/format.h"
#include "cpluslib/format.h"
#include "ui/widgets/quit.h"


/** Add report line. Creates a properly aligned "what: amount unit"
    line on the given document. */
static void
addReportLine(RichDocument& doc, string_t what, int amount, string_t unit)
{
    doc.add(what);
    doc.addRight(font_ems[FONT_NORMAL]*20, numToString(amount));
    if (unit.size()) {
        doc.add(" " + unit);
    }
    doc.addNewline();
}

/** Last known cargo part of report. Writes a report about last known
    cargo ranges and turn number. Does not modify \c doc if no
    information for such a report is known. */
static void
addLastKnownCargo(RichDocument& doc, const GShip& sh)
{
    /* Helper */
    struct Helper {
        RichDocument& doc;
        int turn;
        bool had_heading;
        int32_t total;

        Helper(RichDocument& doc, int turn)
            : doc(doc), turn(turn), had_heading(false), total(0)
            { }

        void add(GCargoType type, int amount)
            {
                if (!had_heading) {
                    doc.add(RichText(_("Last known cargo"))
                            .withAttribute(RichTextStyleAttribute::Big)
                            .withAttribute(RichTextStyleAttribute::Underline));
                    doc.addNewline();
                    if (turn > 0) {
                        doc.add(format("(%s)", formatTurnNumber(turn)));
                    } else {
                        doc.add(_("(age of this information is unknown)"));
                    }
                    doc.addParagraph();
                    had_heading = true;
                }
                addReportLine(doc, getCargoName(type), amount, getCargoUnit(type));
                if (type != el_Money) {
                    total += amount;
                }
            }
    };
    Helper h(doc, sh.getHistoryTimestamp(sts_Rest));

    /* List cargo */
    static const uint8_t list[] = {
        el_Neutronium, el_Tritanium, el_Duranium, el_Molybdenum,
        el_Supplies, el_Colonists, el_Money
    };
    for (std::size_t i = 0; i < countof(list); ++i) {
        GCargoType el = GCargoType(list[i]);
        if (sh.getCargoRaw(el).isKnown()) {
            h.add(el, sh.getCargoRaw(el));
        }
    }
    if (sh.getAmmoRaw().isKnown()) {
        /* we know its ammo */
        if (sh.getHullId().isKnown() && sh.getHull().getNumBays() > 0) {
            /* we know it's a carrier */
            h.add(el_Fighters, sh.getAmmoRaw());
        } else if (sh.getTorpType().isKnown()
                   && sh.getTorpType() > 0 && sh.getTorpType() <= NUM_TORPS
                   && ((sh.getNumTorpLaunchers().isKnown() && sh.getNumTorpLaunchers() > 0)
                       || (sh.getHullId().isKnown() && sh.getHull().getMaxLaunchers() > 0)))
        {
            /* we know its torpedo type, and we know it has launchers */
            h.add(getCargoTypeFromTorpType(sh.getTorpType()), sh.getAmmoRaw());
        } else {
            /* can this happen? PCC 1.x will display a naked "Torpedoes" line here. */
        }
    }

    /* Did we have anything? */
    if (!h.had_heading) {
        return;
    }

    /* Footer */
    doc.addNewline();
    addReportLine(doc, _("\xE2\x96\xB6 Total"), h.total, _("kt"));
    doc.addNewline();
}

/** Mass ranges part or report. Attempts to derive useful information
    from a ship's mass. Does nothing (and does not modify \c doc) when
    the ship's mass or type are not known. */
static void
addMassRanges(RichDocument& doc, const GShip& sh)
{
    /* We need something to work with */
    if (!sh.getHullId().isKnown() || !sh.getMass().isKnown()) {
        return;
    }
    const GHull& h = sh.getHull();

    /* Heading */
    doc.add(RichText(format(_("Current mass: %d kt"), numToString(sh.getMass())))
            .withAttribute(RichTextStyleAttribute::Big)
            .withAttribute(RichTextStyleAttribute::Underline));
    doc.addParagraph();

    /* We know the ship's mass and hull. Do we know its arms? */
    bool armsKnown = true;
    int knownArmsTotal = 0;
    int tubeMass;
    addReportLine(doc, _("Hull"), h.getMass(), _("kt"));
    if (h.getNumBays() > 0) {
        /* It has bays, which don't have a weight */
        tubeMass = 0;
    } else if (sh.getTorpType().isKnown() && sh.getTorpType() > 0
               && sh.getNumTorpLaunchers().isKnown() && sh.getNumTorpLaunchers() > 0)
    {
        /* We know its torpedo type */
        tubeMass = getTorp(sh.getTorpType()).getMass() * sh.getNumTorpLaunchers();
        knownArmsTotal += tubeMass;
        addReportLine(doc, format(_("%s launchers"), getTorp(sh.getTorpType()).getName()), tubeMass, _("kt"));
    } else if (h.getMaxLaunchers() > 0 && !sh.getNumTorpLaunchers().isKnown()) {
        /* It might have torps, but we don't know which one */
        armsKnown = false;
        tubeMass = 0;
        for (int i = 1; i <= NUM_TORPS; ++i)
            tubeMass = std::max(tubeMass, getTorp(i).getMass());
        tubeMass *= h.getMaxLaunchers();
    } else {
        /* No torpedoes */
        tubeMass = 0;
    }

    int beamMass;
    if (sh.getBeamType().isKnown() && sh.getBeamType() > 0
        && sh.getNumBeams().isKnown() && sh.getNumBeams() > 0)
    {
        /* We know its beam type */
        beamMass = getBeam(sh.getBeamType()).getMass() * sh.getNumBeams();
        knownArmsTotal += beamMass;
        addReportLine(doc, format(_("%s beams"), getBeam(sh.getBeamType()).getName()), beamMass, _("kt"));
    } else if (h.getMaxBeams() > 0 && !sh.getNumBeams().isKnown()) {
        /* It might have beams but we don't know which ones */
        armsKnown = false;
        beamMass = 0;
        for (int i = 1; i <= NUM_BEAMS; ++i)
            beamMass = std::max(beamMass, getBeam(i).getMass());
        beamMass *= h.getMaxBeams();
    } else {
        /* No beams */
        beamMass = 0;
    }

    /* Now add derived information. We know that
           Hull_mass   == h.getMass()
           Weapon_mass <= beamMass + tubeMass
           Fuel_mass   <= h.getFuel()
           Cargo_mass  <= h.getCargo()
       Hence,
           Fuel_mass == Mass - Hull_mass - Weapon_mass - Cargo_mass
       and therefore
           Fuel_mass >= Mass - Hull_mass - (beamMass+tubeMass) - h.getCargo()
       giving a range for fuel content. Same reasoning holds for cargo. */
    int minCargo = 0;
    int minFuel = 0;
    int mass = sh.getMass() - h.getMass() - beamMass - tubeMass;
    if (mass > h.getCargo())
        minFuel = mass - h.getCargo();
    if (mass > h.getFuel())
        minCargo = mass - h.getFuel();

    /* Mass is
           Mass := Hull_mass + Weapon_mass + Fuel_mass + Cargo_mass
       We know that
           Hull_mass   == h.getMass()
           Weapon_mass >= knownArmsTotal
           Fuel_mass   >= minFuel
           Cargo_mass  >= minCargo
       Hence,
           Fuel_mass   == Mass - Hull_mass - Weapon_mass - Cargo_mass
           Fuel_mass   <= Mass - Hull_mass - knownArmsTotal - minCargo */
    mass = sh.getMass() - knownArmsTotal - h.getMass();
    addReportLine(doc, armsKnown ? _("Cargo+Fuel") : _("Cargo+Fuel+Weapons"), mass, _("kt"));
    if (mass >= minCargo && mass - minCargo < h.getFuel())
        addReportLine(doc, _("\xE2\x96\xB6 Max. Fuel"), mass - minCargo, _("kt"));
    if (minFuel > 0)
        addReportLine(doc, _("\xE2\x96\xB6 Min. Fuel"), minFuel, _("kt"));
    if (mass >= minFuel && mass - minFuel < h.getCargo())
        addReportLine(doc, armsKnown ? _("\xE2\x96\xB6 Max. Cargo") : _("\xE2\x96\xB6 Max. Cargo+Weapons"), mass - minFuel, _("kt"));
    if (minCargo > 0)
        addReportLine(doc, _("\xE2\x96\xB6 Min. Cargo"), minCargo, _("kt"));
    doc.addNewline();
}

/** Cargo history dialog. Creates a report and displays it, or informs
    the user when the report cannot be created due to lack of
    information. */
void
showCargoHistory(const GShip& sh)
{
    ObjectHolder h;

    const int width = 25*font_ems[FONT_NORMAL]+5;
    UIRichDocument& doc = h.add(new UIRichDocument(GfxPoint(width, 100), 0));
    RichDocument& d = doc.getDocument();
    d.clear();
    d.setPageWidth(width);
    d.setLeftMargin(5);
    addLastKnownCargo(d, sh);
    addMassRanges(d, sh);
    d.finish();

    if (d.getDocumentHeight() == 0) {
        /* Document is empty because we have no information */
        messageBox(_("Nothing is known about this ship's cargo."), _("Ship History"));
    } else {
        /* Build and display window */
        doc.adjustToDocumentSize();

        UIWindow& win = h.add(new UIWindow(_("Ship History"), UIWindow::BLUE_WINDOW, COLOR_GRAY));
        win.setLayout(UIVBoxLayout::instance5);
        win.add(doc);

        UIGroup& g = h.add(new UIGroup(UIHBoxLayout::instance5));
        g.add(h.add(new UISpacer()));
        g.add(h.add(new UIButton(1, _("OK"), SDLK_RETURN)).addStop());
        g.add(h.add(new UISpacer()));
        win.add(g);
        win.add(h.add(new UIKeyTranslator()).addStop(' ', 1).addStop(SDLK_ESCAPE, 1));
        win.add(h.add(new UIQuit(1)));
        win.pack();
        ui_root->centerWidget(win);
        ui_root->runModal(win);
    }
}
