/**
  *  \file game/ion.cc
  *  \brief Ion Storms
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/ion.h"
#include "util/nls.h"
#include "game/univ.h"
#include "cpluslib/format.h"
#include "game/utildata.h"

/** Create ion storm. Initializes the object for a non-existant
    storm. */
GIonStorm::GIonStorm(int id)
    : id(id)
{
    data.volt = 0;              // non-existant
}

GIonStorm::~GIonStorm()
{ }

/** Add data from koreX.dat. */
void
GIonStorm::addKoreData(const TKoreStorm& kstorm)
{
    data = kstorm;
    grow = (kstorm.volt & 1) != 0;
}

/** Add data from utilX.dat.
    \pre ustorm.storm_id == getId() */
void
GIonStorm::addUtilData(const TUtil17Ion& ustorm)
{
    ASSERT(ustorm.storm_id == id);
    data.x       = ustorm.x;
    data.y       = ustorm.y;
    data.volt    = ustorm.volt;
    data.heading = ustorm.heading;
    data.speed   = ustorm.speed;
    data.radius  = ustorm.radius;
    grow         = ustorm.grows != 0;
    // We ignore ustorm.storm_class, it is always computed from voltags
}

string_t
GIonStorm::getName(Name which) const
{
    switch (which) {
     case PlainName:
        return name;
     case LongName:
     case DetailedName:
        return format(_("Ion storm #%d: %s"), id, name);
    }
    return string_t();
}

int
GIonStorm::getId() const
{
    return id;
}

int
GIonStorm::getOwner() const
{
    return 0;
}

GPoint
GIonStorm::getPos() const
{
    return GPoint(data.x, data.y);
}

int
GIonStorm::getRadius() const
{
    return data.radius;
}

int32_t
GIonStorm::getRadiusSquared() const
{
    return int32_t(data.radius) * data.radius;
}

int
GIonStorm::getClass() const
{
    if (data.volt > 200)
        return 5;
    else
        return data.volt / 50 + 1;
}

/***************************** GIonStormType *****************************/

GIonStormType::GIonStormType(GUniverse& univ)
    : GIndexedObjectType(NUM_STORMS),
      univ(univ)
{ }

GIonStorm&
GIonStormType::getObjectByIndex(int index)
{
    return univ.getIonStorm(index);
}

GUniverse*
GIonStormType::getUniverseByIndex(int /*index*/)
{
    return &univ;
}

bool
GIonStormType::isValidIndex(int index) const
{
    return index > 0
        && index <= NUM_STORMS
        && univ.getIonStorm(index).getVoltage() != 0;
}
