/**
  *  \file game/objsel.cc
  *  \brief Object Selection
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/objsel.h"
#include "game/objtypes.h"
#include "cpluslib/assert.h"

/** Destructor. */
GObjectSelection::~GObjectSelection()
{ }

/** Get currently selected object.
    \return pointer to object if any selected; null if none */
GObject*
GObjectSelection::getCurrentObject() const
{
    int index = getCurrentIndex();
    GObjectType* type = getObjectType();
    if (type != 0 && type->isValidIndex(index)) {
        return &type->getObjectByIndex(index);
    } else {
        return 0;
    }
}

/** Get currently selected universe. This is the universe the selected
    object is in.
    \return pointer to universe if any object selected; null if none
    (note that the result can also be null for objects that are not in
    any universe, notably list dividers). */
GUniverse*
GObjectSelection::getCurrentUniverse() const
{
    int index = getCurrentIndex();
    GObjectType* type = getObjectType();
    if (type != 0 && type->isValidIndex(index)) {
        return type->getUniverseByIndex(index);
    } else {
        return 0;
    }
}

/************************* GSimpleObjectSelection ************************/

/** Constructor. */
GSimpleObjectSelection::GSimpleObjectSelection()
    : current(0),
      type(0)
{ }

GSimpleObjectSelection::~GSimpleObjectSelection()
{ }

GObjectType*
GSimpleObjectSelection::getObjectType() const
{
    return type;
}

void
GSimpleObjectSelection::setCurrentIndex(int index)
{
    // Validate
    if (index != 0) {
        ASSERT(type != 0);
        ASSERT(type->isValidIndex(index));
    }

    // Execute
    if (index != current) {
        current = index;
        sig_current_changed.raise();
    }
}

int
GSimpleObjectSelection::getCurrentIndex() const
{
    return current;
}

/** Set underlying object type.
    \param type Type, can be null */
void
GSimpleObjectSelection::setObjectType(GObjectType* type)
{
    if (type != this->type) {
        this->type = type;
        if (type != 0) {
            // We're selecting an object type
            conn_set_changed = type->sig_set_changed.add(this, &GSimpleObjectSelection::onSetChanged);

            // onSetChanged happens to do whatever we want to do here
            onSetChanged(0);
        } else {
            // We're selecting null
            conn_set_changed.disconnect();
            current = 0;
            sig_current_changed.raise();
        }
    }
}

/** Initialize from another GObjectSelection. Uses the other selection's type
    and current index. */
void
GSimpleObjectSelection::initFrom(GObjectSelection& other)
{
    // Copy type
    setObjectType(other.getObjectType());

    // Copy position
    if (int index = other.getCurrentIndex()) {
        if (type != 0 && type->isValidIndex(index)) {
            setCurrentIndex(index);
        }
    }
}

/** Handle changes in the underlying set. This function is called on the
    current type's conn_set_changed signal. It changes the current index
    if it got invalid. */
void
GSimpleObjectSelection::onSetChanged(int hint)
{
    if (type->isValidIndex(current)) {
        // OK, stay
    } else if (type->isValidIndex(hint)) {
        // OK, go to new position
        current = hint;
    } else {
        // Find something new
        current = type->findNextIndexWrap(current, false);
    }
    sig_current_changed.raise();
}

/************************* GProxyObjectSelection *************************/

GProxyObjectSelection::GProxyObjectSelection()
    : selection(0)
{ }

GProxyObjectSelection::~GProxyObjectSelection()
{ }

GObjectType*
GProxyObjectSelection::getObjectType() const
{
    return selection != 0 ? selection->getObjectType() : 0;
}

void
GProxyObjectSelection::setCurrentIndex(int index)
{
    if (selection != 0) {
        selection->setCurrentIndex(index);
    }
}

int
GProxyObjectSelection::getCurrentIndex() const
{
    return selection != 0 ? selection->getCurrentIndex() : 0;
}

/** Change the underlying other selection.
    \param selection New underlying selection. Can be null. */
void
GProxyObjectSelection::setObjectSelection(GObjectSelection* selection)
{
    this->selection = selection;
    if (selection != 0) {
        conn_current_changed = selection->sig_current_changed.add(&sig_current_changed, &Signal0::raise);
    } else {
        conn_current_changed.disconnect();
    }
    sig_current_changed.raise();
}
