/**
  *  \file game/talliance.cc
  *  \brief THost alliance handling
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <ctype.h>
#include "game/talliance.h"
#include "game/cmds.h"
#include "game/struct.h"
#include "util/nls.h"
#include "game/host.h"

const char ALLIANCE_ID[] = "thost.ally";
const char STRONG_ID[] = "thost.ff";


GTHostAllianceHandler::GTHostAllianceHandler(GCommandContainer& commands)
    : commands(commands)
{ }

void
GTHostAllianceHandler::postprocess(GAlliances& allies)
{
    // Find offers
    GAllianceOffer* pAlliance = allies.getMutableOffer(allies.find(ALLIANCE_ID));
    GAllianceOffer* pStrong   = allies.getMutableOffer(allies.find(STRONG_ID));

    // Reset to defaults
    if (pAlliance) {
        pAlliance->newOffer = pAlliance->oldOffer;
    }
    if (pStrong) {
        pStrong->newOffer = pStrong->oldOffer;
    }

    // Check the command messages
    if (const GCommand* cmd = commands.getCommand(GCommand::phc_TAlliance, 0)) {
        const string_t arg = cmd->getArg();
        for (size_t i = 0; i+2 < arg.size(); i += 3) {
            // Determine race
            int player = getPlayerIdFromChar(arg[i+2]);
            if (player > 0) {
                // Set result
                if (pAlliance) {
                    pAlliance->newOffer[player] = (arg[i] == 'e' ? GAllianceOffer::No : GAllianceOffer::Yes);
                }
                if (pStrong) {
                    pStrong->newOffer[player] = (arg[i] == 'F' ? GAllianceOffer::Yes : GAllianceOffer::No);
                }
            }
        }
    }
}

void
GTHostAllianceHandler::handleChanges(const GAlliances& allies)
{
    const GAllianceOffer* pAlliance = allies.getOffer(allies.find(ALLIANCE_ID));
    const GAllianceOffer* pStrong   = allies.getOffer(allies.find(STRONG_ID));

    // Build fcode list
    string_t codes;
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        bool oldAlliance = pAlliance && GAllianceOffer::isOffer(pAlliance->oldOffer[i]);
        bool newAlliance = pAlliance && GAllianceOffer::isOffer(pAlliance->newOffer[i]);
        bool oldStrong   = pStrong   && GAllianceOffer::isOffer(pStrong->oldOffer[i]);
        bool newStrong   = pStrong   && GAllianceOffer::isOffer(pStrong->newOffer[i]);
        
        if (oldAlliance != newAlliance || (newAlliance && oldStrong != newStrong)) {
            char code = (!newAlliance ? 'e' : newStrong ? 'F' : 'f');
            codes += code;
            codes += code;
            codes += std::tolower(getPlayerChar(i));
        }
    }

    // Make the command
    if (codes.empty()) {
        commands.removeCommand(GCommand::phc_TAlliance, 0);
    } else {
        commands.setCommand(GCommand::phc_TAlliance, 0, codes);
    }
}

void
GTHostAllianceHandler::processVersion(GCommandContainer& cmd, GAlliances& alliances, uint32_t version)
{
    // We pretend all host versions have alliances.
    // This is the same as PCC 1.x.
    if (alliances.find(ALLIANCE_ID) == GAlliances::nil) {
        alliances.addLevel(GAllianceLevel(_("Standard alliance"), ALLIANCE_ID, GAllianceLevel::IsOffer | GAllianceLevel::IsCombat));
        if (version == 0 || version >= MKVERSION(3,22,39)) {
            alliances.addLevel(GAllianceLevel(_("Vision alliance"), STRONG_ID, GAllianceLevel::NeedsOffer));
        }
        alliances.addNewHandler(new GTHostAllianceHandler(cmd));
    }
}
