/**
  *  \file game/utilparse.cc
  *  \brief UTIL.DAT parsing for game information
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "cpluslib/format.h"
#include "game/host.h"
#include "game/ion.h"
#include "game/planet.h"
#include "game/ship.h"
#include "game/univ.h"
#include "game/utildata.h"
#include "game/utilparse.h"
#include "util/console.h"
#include "util/nls.h"
#include "game/stat.h"
#include "game/unitscore.h"
#include "game/db.h"
#include "util/recode.h"
#include "io/streams.h"
#include "game/global.h"
#include "game/palliance.h"
#include "cpluslib/misc.h"
#include "game/classicvcr.h"
#include "game/team.h"

static const TUtil7Battle*
getBattleResult(const std::vector<TUtil7Battle>& bs, const GClassicVcrEntry& entry)
{
    // ex phost.pas:GetBattleResult
    for (size_t i = 0, n = bs.size(); i < n; ++i) {
        const TUtil7Battle& cand = bs[i];
        const GClassicVcr& vcr = entry.getVcr();
        if (cand.id[0] == vcr(0).getId()
            && cand.id[1] == vcr(1).getId()
            && ((cand.battle_type != 0) == (vcr(1).isPlanet()))
            && cand.owner[0] == vcr(0).getOwner()
            && cand.owner[1] == vcr(1).getOwner()
            && (cand.seed == 0 || uint16_t(cand.seed) == uint16_t(vcr.getSeed())))
        {
            return &cand;
        }
    }
    return 0;
}

static void
processPlanetResult(GGameTurn& trn, const GClassicVcrEntry& entry, int side, GPoint& pos)
{
    const GClassicVcr& vcr = entry.getVcr();
    const GVcrObject& obj = vcr(side);
    GUniverse& univ = trn.getCurrentUniverse();
    if (obj.isPlanet() && univ.isValidPlanetId(obj.getId())) {
        // Update position if known
        GPlanet& pl = univ.getPlanet(obj.getId());
        if (pl.isVisible()) {
            pos = pl.getPos();
        }

        // Report last-turn ownership
        GMessageInformation info(MessagePlanet, obj.getId(), trn.getTurnNumber()-1);
        info.addValue(mi_Owner, obj.getOwner());
        pl.addMessageInformation(info, -1);
    }
}

static void
processShipResult(GGameTurn& trn, const GClassicVcrEntry& entry, int side, const GPoint& pos,
                  const std::vector<bool>& destroyed_ships,
                  const TUtil7Battle* br,
                  bool useESB)
{
    const GClassicVcr& vcr = entry.getVcr();
    const GVcrObject& obj = vcr(side);
    GUniverse& univ = trn.getCurrentUniverse();
    if (!obj.isPlanet() && univ.isValidShipId(obj.getId())) {
        GShip& sh = univ.getShip(obj.getId());
        if (sh.getOwner() != getPlayerId()) {
            // Prepare some last-turn information
            GMessageInformation info(MessageShip, obj.getId(), trn.getTurnNumber()-1);
            info.addValue(ms_Name, obj.getName());
            if (int hullNr = obj.getGuessedHull()) {
                info.addValue(mi_ShipHull, hullNr);
                if (int engineNr = obj.identifyEngine(hullNr, useESB)) {
                    info.addValue(mi_ShipEngineType, engineNr);
                }
            }
            info.addValue(mi_Owner, obj.getOwner());
            info.addValue(mi_ShipBeamType, obj.getBeamType());
            info.addValue(mi_ShipNumBeams, obj.getBeamCount());
            info.addValue(mi_ShipNumBays, obj.getBayCount());
            info.addValue(mi_ShipLauncherType, obj.getTorpType());
            info.addValue(mi_ShipNumLaunchers, obj.getTorpLauncherCount());
            info.addValue(mi_ShipAmmo, obj.getBayCount() == 0 ? obj.getTorpCount() : obj.getFighterCount());
            info.addValue(mi_Damage, obj.getDamage());
            sh.addMessageInformation(info, -1);

            // Prepare current-turn information
            if (obj.getId() > 0 && obj.getId() < int(destroyed_ships.size()) && !destroyed_ships[obj.getId()]) {
                GMessageInformation newInfo(MessageShip, obj.getId(), trn.getTurnNumber());
                if (pos != GPoint(0, 0)) {
                    newInfo.addValue(mi_X, pos.x);
                    newInfo.addValue(mi_Y, pos.y);
                }
                if (br != 0) {
                    newInfo.addValue(mi_Damage, br->result[side] == 2 ? 999 : br->damage_after[side]);
                    if (obj.getBayCount() != 0 && obj.getFighterCount() != 0) {
                        newInfo.addValue(mi_ShipAmmo, br->fighter_after[side]);
                    }
                    if (obj.getTorpLauncherCount() != 0 && obj.getTorpCount() != 0) {
                        newInfo.addValue(mi_ShipAmmo, br->torp_after[side]);
                    }
                    newInfo.addValue(mi_Owner, br->result[side] == 1 ? br->owner[!side] : br->owner[side]);
                }
                sh.addMessageInformation(newInfo, -1);
            }
        }
    }
}

/** Load score record.
    \param data   Pointer to record content
    \param size   Size of record content
    \param max_id Maximum permitted unit Id
    \param defs   GUnitScoreDefinitions object describing this class's unit scores
    \param get    Function to get a unit's scores from a univers object
    \param univ   Universe */
static void
loadScoreRecord(const char* data,
                std::size_t size,
                const int max_id,
                GUnitScoreDefinitions& defs,
                GUnitScores& get(GUniverse&, int),
                GUniverse& univ)
{
    if (size < TUnitScoreHeader::size)
        return;

    /* Read header */
    TUnitScoreHeader ush;
    getStructure(data, ush);
    data += ush.size;
    size -= ush.size;

    /* Remember header */
    GUnitScoreDefinitions::index_t index = defs.addScoreDefinition(ush);

    /* Read entries */
    while (size >= 4) {
        int16 id    = getInt16(data);
        int16 score = getInt16(data+2);
        data += 4;
        size -= 4;
        if (id > 0 && id <= max_id)
            get(univ, id).mergeScore(index, score, univ.getTurnNumber());
    }
}

/** Constructor.
    \param univ Universe object. May be partially constructed only. Currently we only
                take the turn number from it. */
GUtilMessageParser::GUtilMessageParser(const GUniverse& univ)
    : univ(univ),
      battle_results(),
      destroyed_ships(NUM_SHIPS+1)
{ }

void
GUtilMessageParser::killShip(int sid)
{
    if (sid > 0 && sid < int(destroyed_ships.size())) {
        destroyed_ships[sid] = true;
    }
}

/** Parse a util.dat record.
    \param rec_id [in] Record type
    \param size   [in] Data size
    \param data   [in] Data
    \param info_out [out] Receives parsed information */
void
GUtilMessageParser::parseRecord(const int rec_id, std::size_t size, const char* data,
                                ptr_vector<GMessageInformation>& info_out)
{
    switch (rec_id) {
     case 0:
     case 46:
        /* Minefield */
        if (size >= 14) {
            TUtil0Minefield mf;
            getPartialStructure(data, mf, size);
            GMessageInformation& info =
                *info_out.push_back(new GMessageInformation(MessageMinefield, mf.mine_id, univ.getTurnNumber()));
            info.addValue(mi_X, mf.x);
            info.addValue(mi_Y, mf.y);
            info.addValue(mi_Owner, mf.owner);
            info.addValue(mi_MineUnits, mf.units);
            info.addValue(mi_MineType, mf.type);
            if (size >= 18) {
                info.addValue(mi_MineScanReason, mf.cause+1);
            }
        }
        break;

     case 1:
        /* Explosion */
        if (size >= 6) {
            /* FIXME: This only updates the ship. We must generate a bang marker, too. */
            TUtil1Bang bang;
            getPartialStructure(data, bang, size);
            GMessageInformation& info =
                *info_out.push_back(new GMessageInformation(MessageShip, bang.sid, univ.getTurnNumber()));
            info.addValue(mi_Damage, 999);
            info.addValue(mi_X, bang.x);
            info.addValue(mi_Y, bang.y);
            if (size >= 26) {
                info.addValue(ms_Name, convertGameToUtf8(getBasicString(bang.name)));
            }
            killShip(bang.sid);
        }
        break;

     case 2:
        /* Mine hit */
        if (size >= 8) {
            TUtil2MineHit hit;
            getPartialStructure(data, hit, size);
            GMessageInformation& info =
                *info_out.push_back(new GMessageInformation(MessageShip, hit.sid, univ.getTurnNumber()));
            // Do NOT generate a position.
            // This would cause unreliable targets to be generated (e.g., ship moves on after hit),
            // and - worse - possibly damage more precise guesses.
            //   info.addValue(mi_X, hit.x);
            //   info.addValue(mi_Y, hit.y);
            info.addValue(mi_Damage, hit.damage);
            if (size >= 28) {
                info.addValue(ms_Name, convertGameToUtf8(getBasicString(hit.name)));
            }
            if (hit.damage >= 100) {
                // This is PHost; anything over 100 kills the ship
                killShip(hit.sid);
            }
        }
        break;

     case 3:
        /* Dark Sense */
        if (size >= 26) {
            TUtil3DarkSense report;
            getPartialStructure(data, report, size);
            GMessageInformation& info =
                *info_out.push_back(new GMessageInformation(MessagePlanet, report.pid, univ.getTurnNumber()));
            info.addValue(mi_Owner, report.owner);
            info.addValue(mi_PlanetTotalN, report.ore[0]);
            info.addValue(mi_PlanetTotalT, report.ore[1]);
            info.addValue(mi_PlanetTotalD, report.ore[2]);
            info.addValue(mi_PlanetTotalM, report.ore[3]);
            info.addValue(mi_PlanetCash, report.money);
            info.addValue(mi_PlanetHasBase, report.starbase);
        }
        break;

     case 4:
        /* Super Spy */
        if (size >= 31) {
            TUtil4SuperSpy report;
            getPartialStructure(data, report, size);
            GMessageInformation& info =
                *info_out.push_back(new GMessageInformation(MessagePlanet, report.pid, univ.getTurnNumber()));
            info.addValue(mi_PlanetMines, report.mines);
            info.addValue(mi_PlanetFactories, report.factories);
            info.addValue(mi_PlanetDefense, report.defense);
            info.addValue(ms_FCode, convertGameToUtf8(getBasicString(report.fcode)));
            info.addValue(mi_PlanetTotalN, report.ore[0]);
            info.addValue(mi_PlanetTotalT, report.ore[1]);
            info.addValue(mi_PlanetTotalD, report.ore[2]);
            info.addValue(mi_PlanetTotalM, report.ore[3]);
            info.addValue(mi_PlanetCash, report.money);
            if (size >= 35) {
                info.addValue(mi_PlanetSupplies, report.supplies);
            }
        }
        break;

     case 5:
        /* Planet Exploration */
        if (size >= 12) {
            TUtil5Planet report;
            getPartialStructure(data, report, size);
            GMessageInformation& info =
                *info_out.push_back(new GMessageInformation(MessagePlanet, report.pid, univ.getTurnNumber()));
            info.addValue(mi_PlanetTemperature, report.temp_real);
            info.addValue(mi_Owner, report.owner);
            info.addValue(mi_PlanetColonists, report.colonists_real / 100);
            info.addValue(mi_PlanetHasBase, report.starbase);
        }
        break;

     case 6:
        /* Sensor Sweep */
        if (size >= 6) {
            TUtil6SensorSweep report;
            getPartialStructure(data, report, size);
            GMessageInformation& info =
                *info_out.push_back(new GMessageInformation(MessagePlanet, report.pid, univ.getTurnNumber()));
            info.addValue(mi_Owner, report.owner);
            info.addValue(mi_PlanetActivity, report.activity);
        }
        break;

     case 7:
        if (size >= 26) {
            TUtil7Battle report;
            report.seed = 0;
            report.x = 0;
            report.y = 0;
            getPartialStructure(data, report, size);
            battle_results.push_back(report);

            // If either ship blew up, mark so.
            if (report.result[0] == 2) {
                killShip(report.id[0]);
            }
            if (report.battle_type == 0 && report.result[1] == 2) {
                killShip(report.id[1]);
            }
        }
        break;

     case 8:
        /* Meteor */
        if (size >= 18) {
            TUtil8Meteor report;
            getPartialStructure(data, report, size);
            GMessageInformation& info =
                *info_out.push_back(new GMessageInformation(MessagePlanet, report.pid, univ.getTurnNumber()));
            info.addValue(mi_PlanetAddedN, report.ore[0]);
            info.addValue(mi_PlanetAddedT, report.ore[1]);
            info.addValue(mi_PlanetAddedD, report.ore[2]);
            info.addValue(mi_PlanetAddedM, report.ore[3]);
        }
        break;

     case 9:
        /* Meteorite */
        /* This could be handled like a Meteor, but does not provide useful information:
           it is only sent to the planet owner who knows the planet's content anyway,
           and even if we had these reports for foreign planets, the added amounts are
           usually minor compared to the existing amounts. Hence we ignore it. */
        break;

        /* TUtil10Target handled outside */

     case 11:
        /* Allied base */
        if (size >= 4) {
            TUtil11AlliedBase report;
            getPartialStructure(data, report, size);
            GMessageInformation& info =
                *info_out.push_back(new GMessageInformation(MessagePlanet, report.bid, univ.getTurnNumber()));
            info.addValue(mi_Owner, report.owner);
            info.addValue(mi_PlanetHasBase, 1);
        }
        break;

     case 12:
        /* Allied planet */
        if (size >= 42) {
            TUtil12AlliedPlanet report;
            getPartialStructure(data, report, size);
            GMessageInformation& info =
                *info_out.push_back(new GMessageInformation(MessagePlanet, report.pid, univ.getTurnNumber()));
            info.addValue(mi_Owner, report.owner);
            info.addValue(mi_PlanetTemperature, report.temp_real);
            info.addValue(mi_PlanetNativeRace, report.native_race);
            info.addValue(mi_PlanetNativeGov, report.native_gov);
            info.addValue(mi_PlanetNatives, report.natives_real / 100);
            info.addValue(mi_PlanetMinedN, report.ore_mined[0]);
            info.addValue(mi_PlanetMinedT, report.ore_mined[1]);
            info.addValue(mi_PlanetMinedD, report.ore_mined[2]);
            info.addValue(mi_PlanetMinedM, report.ore_mined[3]);
            info.addValue(mi_PlanetColonists, report.colonists_real / 100);
            info.addValue(mi_PlanetSupplies, report.supplies);
            info.addValue(mi_PlanetCash, report.money);
        }
        break;

        /* TUtil13Control handled outside */

     case 15:
        /* Wormhole travel */
        if (size >= 12) {
            /* The only useful information we get from this is the ship's new damage,
               which we may not know if the ship is under remote control */
            TUtil15WormholeTravel report;
            getPartialStructure(data, report, size);
            GMessageInformation& info =
                *info_out.push_back(new GMessageInformation(MessageShip, report.sid, univ.getTurnNumber()));
            info.addValue(mi_Damage, report.damage);
        }
        break;

     case 16:        /* Ship recycled */
     case 18:        /* Ship colonized */
        if (size >= 4) {
            TUtil15WormholeTravel report;
            getPartialStructure(data, report, size);
            GMessageInformation& info =
                *info_out.push_back(new GMessageInformation(MessageShip, report.sid, univ.getTurnNumber()));
            info.addValue(mi_Damage, 999);
            killShip(report.sid);
        }
        break;

        /* TUtil17Ion handled outside */

     case 19:
        /* Ship surrendered */
        if (size >= 8) {
            /* This yields two parts of information:
               - ship /sid/ now belongs to /new_owner/
               - /new_owner/ has a base on planet /bid/ */
            TUtil19Surrender report;
            getPartialStructure(data, report, size);

            GMessageInformation& ship_info =
                *info_out.push_back(new GMessageInformation(MessageShip, report.sid, univ.getTurnNumber()));
            ship_info.addValue(mi_Owner, report.new_owner);

            GMessageInformation& planet_info =
                *info_out.push_back(new GMessageInformation(MessagePlanet, report.bid, univ.getTurnNumber()));
            planet_info.addValue(mi_Owner, report.new_owner);
            planet_info.addValue(mi_PlanetHasBase, 1);
        }
        break;

        // TODO: TUtil20ShipBuilt. Can we use it?

     case 21:
        /* Ship trade */
        if (size >= 6) {
            TUtil21ShipGiven report;
            getPartialStructure(data, report, size);

            GMessageInformation& info =
                *info_out.push_back(new GMessageInformation(MessageShip, report.sid, univ.getTurnNumber()));
            info.addValue(mi_Owner, report.new_owner);
        }
        break;

     case 23:
        /* Bioscan */
        if (size >= 10) {
            TUtil23Bioscan report;
            getPartialStructure(data, report, size);
            GMessageInformation& info =
                *info_out.push_back(new GMessageInformation(MessagePlanet, report.pid, univ.getTurnNumber()));
            info.addValue(mi_PlanetNativeRace, report.native_race);
            info.addValue(mi_PlanetNatives, report.native_real / 100);
            info.addValue(mi_PlanetTemperature, report.temp_real);
        }
        break;

     case 24:
        /* Glory device */
        if (size >= 6) {
            /* FIXME: right now, this only generates information that the ship
               exploded. We should also generate a marker? */
            TUtil24GD report;
            getPartialStructure(data, report, size);
            GMessageInformation& info =
                *info_out.push_back(new GMessageInformation(MessageShip, report.sid, univ.getTurnNumber()));
            info.addValue(mi_Damage, 999);
            killShip(report.sid);
        }
        break;

     case 25:
        /* Glory damage */
        if (size >= 10) {
            TUtil25GDHit report;
            getPartialStructure(data, report, size);
            GMessageInformation& info =
                *info_out.push_back(new GMessageInformation(MessageShip, report.sid, univ.getTurnNumber()));
            info.addValue(mi_X, report.x);
            info.addValue(mi_Y, report.y);
            info.addValue(mi_Damage, report.damage);
            info.addValue(mi_Owner, report.owner);
            if (size >= 12)
                info.addValue(mi_ShipHull, report.hull);
            if (size >= 32)
                info.addValue(ms_Name, convertGameToUtf8(getBasicString(report.name)));
            if (report.damage >= 100) {
                killShip(report.sid);
            }
        }
        break;

     case 26:
        /* Boarding */
        if (size >= 6) {
            /* This generates information about the new owner of /sid/, and,
               if present, the owner of /boarder_sid/. */
            TUtil26Boarded report;
            getPartialStructure(data, report, size);
            GMessageInformation& info =
                *info_out.push_back(new GMessageInformation(MessageShip, report.sid, univ.getTurnNumber()));
            info.addValue(mi_Owner, report.new_owner);

            if (size >= 8) {
                GMessageInformation& boarder_info =
                    *info_out.push_back(new GMessageInformation(MessageShip, report.boarder_sid, univ.getTurnNumber()));
                boarder_info.addValue(mi_Owner, report.new_owner);
            }
        }
        break;

        /* case 27 (old FTP) handled by GAttachmentUnpacker */

     case 28:
        /* Ground attack */
        if (size >= 8) {
            TUtil28GroundCombat report;
            getPartialStructure(data, report, size);
            GMessageInformation& info =
                *info_out.push_back(new GMessageInformation(MessagePlanet, report.pid, univ.getTurnNumber()));
            info.addValue(mi_Owner,
                          report.result == 0
                          ? report.owner
                          : report.result == 1
                          ? report.attacker
                          : 0);
        }
        break;

     case 29:
        /* Mines destroy mines */
        if (size >= 16) {
            /* This information cannot generate new minefield scans, but it can
               invalidate old ones (coordinate mismatch) or modify existing ones */
            TUtil29MinesExplode report;
            getPartialStructure(data, report, size);

            GMessageInformation& info1 =
                *info_out.push_back(new GMessageInformation(MessageMinefield, report.id1, univ.getTurnNumber()));
            info1.addValue(mi_X, report.x1);
            info1.addValue(mi_Y, report.y1);
            info1.addValue(mi_MineUnitsRemoved, report.exploded_units);

            GMessageInformation& info2 =
                *info_out.push_back(new GMessageInformation(MessageMinefield, report.id2, univ.getTurnNumber()));
            info2.addValue(mi_X, report.x2);
            info2.addValue(mi_Y, report.y2);
            info2.addValue(mi_MineUnitsRemoved, report.exploded_units);
        }
        break;

     case 30:
        /* EOF */
        /* ignored */
        break;

     case 31:
        /* Mine scoop */
        if (size >= 10) {
            TUtil31MineScoop report;
            getPartialStructure(data, report, size);
            GMessageInformation& info =
                *info_out.push_back(new GMessageInformation(MessageMinefield, report.mine_id, univ.getTurnNumber()));
            info.addValue(mi_MineUnitsRemoved, report.mines_swept);
            if (size >= 14)
                info.addValue(mi_MineUnits, report.mines_orig - report.mines_swept);
        }
        break;

     case 32:
        /* Pillage */
        if (size >= 10) {
            /* Recent PHost includes the ship owner, but we cannot do anything
               with that information (useful for player diplomacy, but not history) */
            TUtil32Pillage report;
            getPartialStructure(data, report, size);
            GMessageInformation& info =
                *info_out.push_back(new GMessageInformation(MessagePlanet, report.pid, univ.getTurnNumber()));
            info.addValue(mi_PlanetColonists, report.colonists);
            info.addValue(mi_PlanetNatives, report.natives);
        }
        break;

        /* case 33 General Object handled outside */
        /* case 34 General file handled in GAttachmentUnpacker */

     case 35:
        /* Cloak failure. Can we do anything? */
        break;

     case 36:
        /* Loki decloak */
        if (size >= 8) {
            TUtil36Loki report;
            getPartialStructure(data, report, size);
            GMessageInformation& info =
                *info_out.push_back(new GMessageInformation(MessageShip, report.sid, univ.getTurnNumber()));
            info.addValue(mi_X, report.x);
            info.addValue(mi_Y, report.y);
            info.addValue(mi_Owner, report.owner);
            /* FIXME: what to do with 'before_movement'? */
        }
        break;

     case 37:
        /* Remote */
        while (size >= 4) {
            int16_t sid  = getInt16(data);
            int16_t flag = getInt16(data+2);
            GMessageInformation& info =
                *info_out.push_back(new GMessageInformation(MessageShip, sid, univ.getTurnNumber()));
            info.addValue(mi_ShipRemoteFlag, flag);
            data += 4;
            size -= 4;
        }
        break;

        /* case 38 PAL handled outside */

     case 39:
        /* Build queue entry */
        while (size >= TUtil39Queue::size) {
            TUtil39Queue entry;
            getStructure(data, entry);
            GMessageInformation& info =
                *info_out.push_back(new GMessageInformation(MessagePlanet, entry.bid, univ.getTurnNumber()));
            info.addValue(mi_BaseQueuePos, entry.queue_pos);
            info.addValue(mi_BaseQueuePriority, entry.priority);
            size -= TUtil39Queue::size;
            data += TUtil39Queue::size;
        }
        break;

     case 40:
        /* Web drain complete */
        if (size >= 2) {
            /* Web drain complete reports that the ship is out of fuel.
               Is it useful this way? What about ramscoopers? */
            TUtil40WebDrainComplete report;
            getPartialStructure(data, report, size);
            GMessageInformation& info =
                *info_out.push_back(new GMessageInformation(MessageShip, report.sid, univ.getTurnNumber()));
            info.addValue(mi_ShipFuel, 0);
        }
        break;

     case 41:
        /* RGA */
        if (size >= 4) {
            TUtil41RGA report;
            getPartialStructure(data, report, size);
            GMessageInformation& info =
                *info_out.push_back(new GMessageInformation(MessagePlanet, report.pid, univ.getTurnNumber()));
            info.addValue(mi_PlanetHasNatives, report.has_natives);
        }
        break;

        // TODO: TUtil42GODestroyed

        /* case 43 minefield quotas handled outside */

        // TODO: TUtil44Failure

     case 45:
        /* Planet trade */
        if (size >= 6) {
            TUtil45PlanetGiven report;
            getPartialStructure(data, report, size);

            GMessageInformation& info =
                *info_out.push_back(new GMessageInformation(MessagePlanet, report.sid, univ.getTurnNumber()));
            info.addValue(mi_Owner, report.new_owner);
        }
        break;

        /* case 46 see above */

        /* case 47 Non-existant planets handled outside */

        /* case 48 PAL summary handled outside */

        /* case 49,50 unit score handled outside */

        /* case 51 player score handled outside */

        /* case 52 ship abilities handled outside */

     case 53:
        /* Single minefield explosion */
        if (size >= 10) {
            TUtil53OneMineExplodes report;
            getPartialStructure(data, report, size);

            GMessageInformation& info =
                *info_out.push_back(new GMessageInformation(MessageMinefield, report.mine_id, univ.getTurnNumber()));
            info.addValue(mi_X, report.x);
            info.addValue(mi_Y, report.y);
            info.addValue(mi_MineUnitsRemoved, report.units);
        }
        break;

     case 55:
        /* Production report. Can we do anything with these? */
        break;

     case 56:
        /* Repair report. Can we do anything with these? */
        break;

        /* case 57 ship special handled outside */

     default:
        break;
    }
}

void
GUtilMessageParser::finish(GGameTurn& trn)
{
    // ex ccinit.pas:ProcessBattleResults
    // This is called after loading VCRs and UTIL.DAT, but before loading FLAK.
    // Thus it will not process FLAK results.
    // However, this is also called only when there is no UTIL.DAT.
    GClassicVcrDatabase* db = dynamic_cast<GClassicVcrDatabase*>(&trn.getVcrs());
    if (db == 0) {
        return;
    }

    for (int i = 0, n = db->getNumBattles(); i < n; ++i) {
        GClassicVcrEntry& entry = db->getBattle(i);
        GPoint pos(0, 0);
        const TUtil7Battle* br = getBattleResult(battle_results, entry);
        if (br != 0) {
            if (br->x != 0 && br->y != 0) {
                pos = GPoint(br->x, br->y);
            }
        }

        // Try to process results. Planets first because these may produce a position for later.
        processPlanetResult(trn, entry, 0, pos);
        processPlanetResult(trn, entry, 1, pos);

        bool useESB = config.AllowEngineShieldBonus()
            && (!entry.getVcr()(1).isPlanet() || config.AllowESBonusAgainstPlanets());
        processShipResult(trn, entry, 0, pos, destroyed_ships, br, useESB);
        processShipResult(trn, entry, 1, pos, destroyed_ships, br, useESB);

        // {... add experience levels ...}
        // { FIXME: maybe add some heuristics w.r.t. Commander ships? }
        // IF (pconf<>NIL) AND (pconf^.main.NumExperienceLevels > 0) THEN BEGIN
        //   FOR side:=Left TO Right DO BEGIN
        //     SetStr(uhdr.name, 50, 'Experience');
        //     uhdr.id := uscore_Experience;
        //     uhdr.limit := pconf^.main.NumExperienceLevels;
        //     ush := AddUnitScore((side = Left) OR (VCRs^[i].BattleType = 0), uhdr, 1);
        //     AddUnitScoreEntry(ush, VCRs^[i].Objs[side].Id, VCRs^[i].Objs[side].Level, Gen.TurnNr-1);
        //   END;
        // END;
    }
}

/****************************** GUtilParser ******************************/

/** Constructor.
    \param trn Game turn */
GUtilParser::GUtilParser(GGameTurn& trn, int player)
    : trn(trn), parser(trn.getCurrentUniverse()), player(player)
{ }

GUtilParser::~GUtilParser()
{ }

bool
GUtilParser::process(int rec_id, std::size_t size, const char data[])
{
    GUniverse& univ = trn.getCurrentUniverse();
    switch (rec_id) {
     case 10:
        /* Target */
        if (size >= TShipTarget::size) {
            TShipTarget target;
            getStructure(data, target);
            if (!univ.isValidShipId(target.ship_id)) {
                console.write(LOG_ERROR,
                              format(_("Invalid ship Id #%d for visual contact. Target will be ignored."), target.ship_id));
            } else {
                univ.getShip(target.ship_id).addTargetData(target, GPlayerSet(player));
            }
        }
        break;

     case 13:
        /* Control */
        if (size >= TUtil13ControlMinimal::size) {
            TUtil13Control control;
            getPartialStructure(data, control, size);
            if (univ.getTimestamp() != control.base.timestamp || univ.getTurnNumber() != control.base.turn) {
                console.write(LOG_ERROR, _("util.dat is from a different turn. File will be ignored."));
                return false;
            }
            if (player != control.base.player_id) {
                console.write(LOG_ERROR, _("util.dat belongs to a different player. File will be ignored."));
                return false;
            }

            /* Figure out host version. Encoded as 301002 for 3.1b */
            uint32 version = control.base.phost_major * 100000L + control.base.phost_minor * 1000L;
            if (size >= 89)
                version += control.phost_release & 0x1F;
            host.setHostVersion(version);
            GPHostAllianceHandler::processVersion(trn.getCommands(player), trn.getAlliances(player), version);

            /* Remaining items, not yet checked:
               - spec digests
               - game name */
        } else {
            console.write(LOG_ERROR, _("util.dat control record too short. File is possibly damaged and will be ignored."));
            return false;
        }
        break;

     case 14:
        /* Wormhole. 4.0e adds new fields to the wormhole structure, so we must
           also accept structures where they are missing. */
        // FIXME: it would be cleaner to parse this message into our regular format,
        // and have the Ufo container eat that.
        if (size >= 10) {
            TUtil14Wormhole wh;
            getPartialStructure(data, wh, size);
            if (size < 12)
                wh.ufo_id = -1;
            if (size < 14)
                wh.bidir = -1;
            univ.ty_ufos.addWormholeData(wh);
        }
        break;

     case 17:
        /* Ion storm */
        if (size >= TUtil17Ion::size) {
            TUtil17Ion storm;
            getStructure(data, storm);
            if (univ.isValidIonStormId(storm.storm_id))
                univ.getIonStorm(storm.storm_id).addUtilData(storm);
        }
        break;

     case 22:
        /* Alliances */
        if (size >= 22) {
            TUtil22Alliance allies;
            zeroFill(allies);
            getPartialStructure(data, allies, size);
            GPHostAllianceHandler::processAlliances(trn.getAlliances(player), allies);
        }
        break;

     case 33:
        /* General Object */
        if (size >= TUtil33GO::size) {
            TUtil33GO obj;
            getStructure(data, obj);
            univ.ty_ufos.addObjectData(obj);
        }
        break;

     case 38:
        /* PAL */
        if (size >= TUtil38PAL::size && isStatisticsFileLoaded()) {
            TUtil38PAL pal;
            getStructure(data, pal);

            GStatFile& stat = getStatistics();
            GStatRecord& r = stat.getRecordForTurn(univ.getTurnNumber(), univ.getTimestamp());
            GStatIndex index = stat.getAddIndexForScore(ScoreId_BuildPoints);

            r(index, player) = pal.total;
        }
        break;

     case 43:
        /* Minefield quotas */
        if (size >= 11*2*2 && isStatisticsFileLoaded()) {
            GStatFile& stat = getStatistics();
            GStatRecord& r = stat.getRecordForTurn(univ.getTurnNumber(), univ.getTimestamp());
            GStatIndex quota_index = stat.getAddIndexForScore(ScoreId_MinesAllowed);
            GStatIndex used_index = stat.getAddIndexForScore(ScoreId_MinesLaid);

            for (int pl = 1; pl <= NUM_PLAYERS; ++pl) {
                int16 q = getInt16(data + 2*(pl+1));
                int16 u = getInt16(data + 2*(pl+10));
                r(quota_index, pl) = q;
                if (u >= 0)
                    r(used_index, pl) = u;
            }
        }
        break;

     case 47:
        /* Non-existant planets */
        for (std::size_t i = 0, count = size/2; i < count; ++i) {
            int16_t pid = getInt16(data, 2*i);
            if (univ.isValidPlanetId(pid))
                univ.getPlanet(pid).setKnownToNotExist(true);
        }
        break;

     case 48:
        /* PAL summary */
        if (size >= 11*4 && isStatisticsFileLoaded()) {
            GStatFile& stat = getStatistics();
            GStatRecord& r = stat.getRecordForTurn(univ.getTurnNumber(), univ.getTimestamp());
            GStatIndex index = stat.getAddIndexForScore(ScoreId_BuildPoints);

            for (int pl = 1; pl <= NUM_PLAYERS; ++pl)
                r(index, pl) = getInt32(data + 4*(pl-1));
        }
        break;

     case 49:
        /* Ship score */
        loadScoreRecord(data, size, NUM_SHIPS, ship_score_definitions, getShipScores, univ);
        break;

     case 50:
        /* Planet score */
        loadScoreRecord(data, size, NUM_PLANETS, planet_score_definitions, getPlanetScores, univ);
        break;

     case 51:
        /* Player score */
        if (size >= TUtil51PlayerScore::size && isStatisticsFileLoaded()) {
            TUtil51PlayerScore sc;
            getStructure(data, sc);

            /* Add score to database */
            GStatFile& stat = getStatistics();
            GStatRecord& r = stat.getRecordForTurn(univ.getTurnNumber(), univ.getTimestamp());
            GStatIndex index = stat.getAddIndexForScore(sc.header.score_type);
            for (int pl = 1; pl <= NUM_PLAYERS; ++pl) {
                if (sc.scores[pl-1] >= 0)
                    r(index, pl) = sc.scores[pl-1];
            }

            /* Add description */
            stat.addScoreDescription(sc.header);
        }
        break;

     case 52:
        /* Special functions assigned to ship */
        if (size >= 4) {
            int16_t sid = getInt16(data);
            if (univ.isValidShipId(sid)) {
                for (uint_t i = 1; i < size/2; ++i) {
                    // FIXME: shouldn't this map the Id through getIdFromHostId?
                    univ.getShip(sid).addShipFunction(getInt16(&data[2*i]));
                }
            }
        }
        break;

     case 54:
        /* Enemies */
        if (size >= 2) {
            uint16_t enemies = getUint16(data);
            GPHostAllianceHandler::processEnemies(trn.getAlliances(player), enemies);
        }
        break;

     case 57:
        /* Special function definition */
        if (size >= TUtil57Special::size) {
            TUtil57Special util_func;
            getStructure(data, util_func);

            GHullFunction func(util_func.basic_id, GExpLevelSet::fromInteger(util_func.exp_mask));
            func.setHostId(util_func.func_id);
            GHullFunctionData::instance.getIdFromFunction(func);
        }
        break;

     default:
        // Parse it
        ptr_vector<GMessageInformation> info;
        parser.parseRecord(rec_id, size, data, info);

        // Apply it to universe
        for (ptr_vector<GMessageInformation>::iterator i = info.begin(); i != info.end(); ++i)
            univ.addMessageInformation(**i, -1);
        break;
    }
    return true;
}

void
GUtilParser::failure(Stream& s)
{
    // same message as FileTooShortException
    console.write(LOG_WARN, s.getName() + ": " + _("File is too short"));
}

void
GUtilParser::finish()
{
    parser.finish(trn);
}
