/**
  *  \file reshack/load.cc
  *  \brief Resource Editor: Additional Loaders
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstdio>
#include <memory>
#include <cstring>
#include "reshack/load.h"
#include "gfx/bmfont.h"
#include "io/streams.h"
#include "io/textfile.h"
#include "util/misc.h"


/** Load BDF font. BDF is Adobe's Glyph Bitmap Distribution File
    Format, which is used (among others) as portable storage for X11
    fonts.

    This loader is only in the resource editor, not in the main
    program, because most BDF fonts need a little manual work before
    being actually usable in PCC2.

    This file format doesn't seem to be fully documented anywhere; a
    partial spec is on
    http://partners.adobe.com/public/developer/en/font/5005.BDF_Spec.pdf

    \param font [out] Produce output here
    \param s [in] File to read from
    \return true if any characters were generated */
bool
loadBDF(GfxBitmapFont& font, Stream& s)
{
    int fontWidth = 0;
    int fontHeight = 0;
    int fontBaseline = 0;

    int charWidth = 0;
    int charHeight = 0;
    int charBaseline = 0;

    long charCode = -1;
    bool did = false;

    TextFile tf(s);
    string_t line;
    while (tf.getLine(line)) {
        string_t arg;
        strSplit(line, line, arg, " ");
        if (line == "STARTCHAR" || line == "ENDCHAR") {
            /* STARTCHAR starts a new character. ENDCHAR ends a
               character, in case we've missed it. Both mean we have
               to forget properties of the previous character. */
            charWidth = fontWidth;
            charHeight = fontHeight;
            charBaseline = fontBaseline;
            charCode = -1;
        } else if (line == "ENCODING") {
            /* ENCODING n specifies the codepoint. We assume anything
               in Latin-1 or Unicode so far; users can remap using
               the font editor's remap function. */
            stringValue(arg, charCode);
        } else if (line == "BBX") {
            /* BBX x y dx dy specifies the character's bounding box. 
               We interpret this as:
               - x is the actual glyph width
               - y is the bitmap height, which may be smaller than
                 the actual font height (a property which we cannot
                 handle, all our glyphs have the same height)
               - dy is the origin within the character cell, relative
                 to the font bbox. */
            int x, y, dx, dy;
            if (std::sscanf(arg.c_str(), "%d %d %d %d", &x, &y, &dx, &dy) == 4) {
                charWidth = x;
                charHeight = y;
                charBaseline = dy;
            }
        } else if (line == "FONTBOUNDINGBOX") {
            /* FONTBOUNDINGBOX x y dx dy specifies the font bounding
               box.
               - x just serves as a default width
               - y is the font height
               - dy is the default origin within the character cell */
            int x, y, dx, dy;
            if (std::sscanf(arg.c_str(), "%d %d %d %d", &x, &y, &dx, &dy) == 4) {
                fontWidth = x;
                fontHeight = y;
                fontBaseline = dy;
            }
        } else if (line == "BITMAP") {
            /* BITMAP starts the bitmap definition, one line per pixel
               line, MSB-first hex digits. */
            if (charCode >= 0 && charCode < 0x10000 && charWidth > 0 && charHeight > 0) {
                std::auto_ptr<GfxBitmapGlyph> g(new GfxBitmapGlyph(charWidth, fontHeight));
                for (int y = 0; y < charHeight; ++y) {
                    // Read a line
                    if (!tf.getLine(line))
                        break;
                    if (line == "ENDCHAR")
                        break;

                    // Parse it
                    int x = 0;
                    string_t::size_type n = 0;
                    while (n < line.size() && x < charWidth) {
                        // Read a nibble
                        using std::toupper;
                        const char* dig = "0123456789ABCDEF";
                        const char* p = std::strchr(dig, toupper(uint8_t(line[n])));
                        if (p == 0)
                            break;
                        const int ch = p - dig;
                        // Read four bits
                        for (int i = 0; i < 4 && x < charWidth; ++i) {
                            if (((ch << i) & 8) != 0) {
                                int rx = x;
                                int ry = y + (fontHeight - charHeight) + (fontBaseline - charBaseline);
                                if (rx >= 0 && ry >= 0 && rx < g->getWidth() && ry < g->getHeight()) {
                                    g->set(rx, ry, true);
                                }
                            }
                            ++x;
                        }
                        ++n;
                    }
                }
                font.addNewGlyph(charCode, g.release());
                did = true;
            }
            charCode = -1;
        } else {
            // ignore
        }
    }
    return did;
}
